"""gns3_worker.py

Creates project and nodes on GNS3 server and then populates their configuration."""

import asyncio
from io import BytesIO
import tarfile
import requests
import aiohttp
import docker


def invoker(servername_in: str, gns3_url_in: str, sw_vals_in: list,
            allconf_in: list, prj_id_in: str, connx_in: list):
    """Add nodes to the new GNS3 project and push a copy of the configuration files
    to their substrate docker containers.  Use asyncio/aoihttp to let post requests
    with long completion time run in the background usign cooperative multitasking

    Parameters
    ----------
    servername_in : str
        The name of the aiohttp.ClientSession object used for the connections
    gns3_url_in : str
        The URL to be posted to the GNS3 server
    sw_vals_in : list
        List of needed data about the switches to be emulated
    allconf_in : list
        List-of-lists holding all of the switch's configurations
    connx_in : list
        List of connections we need to make between the GNS3 nodes we're creating
    """

    # Manage an event loop for all of the work done by gns3_node_create_async
    print('')
    print('Creating cEOS nodes in GNS3 project and pushing startup configs to each.')
    sw_vals_new = asyncio.run(gns3_nodes_create_async
                              (servername_in, gns3_url_in, sw_vals_in, allconf_in,
                               prj_id_in))
    # Only AFTER gns3_node_create_async is done, do we start populating connections
    lastwords = asyncio.run(gns3_connx_create_async(servername_in, gns3_url_in,
                                                    sw_vals_new, connx_in, prj_id_in))
    return lastwords


async def gns3_nodes_create_async(servername_in: str, gns3_url_in: str, sw_vals_in:
                                  list, allconf_in: list, prj_id_in: str):
    """Add nodes to the new GNS3 project and push a copy of the configuration files
    to their substrate docker containers.

    Parameters
    ----------
    servername_in : str
        The name of the aiohttp.ClientSession object used for the connections
    gns3_url_in : str
        The URL to be posted to the GNS3 server
    sw_vals_in : list
        List of needed data about the switches to be emulated
    all_conf_in : list
        List-of-lists holding all of the switch's configurations
    """
    print('')
    print('Creating the nodes in the GNS3 project.')
    async with aiohttp.ClientSession() as session:
        # Set x/y coordinates for the first node on the project
        nodex = -825
        nodey = -375
        # Create docker client for RESTful API
        d_clnt = docker.DockerClient(base_url='tcp://'+servername_in+':2375')
        # Initialize single string version of switch config from allconf_in
        my_string_to_go = ''
        tasks = []

        # Loop through the switches and create them in the GNS3 project
        for sw_val_ctr, sw_val in enumerate(sw_vals_in):
            looped_template_id = requests.post(gns3_url_in + 'templates/' + sw_val[7] + '/duplicate', timeout=30).json()['template_id']
            # Put request to change the # of interfaces of the temporary template
            requests.put(gns3_url_in + 'templates/' + looped_template_id,
                         json={'adapters': int(sw_val[6])+1}, timeout=30)
            # Request to instantiate a new node using the temporary template
            newnodeoutput = requests.post(gns3_url_in + 'projects/' + prj_id_in +
                                          '/templates/' + looped_template_id, timeout=30,
                                          json={'x': nodex, 'y': nodey})
            # Capture the GNS3 node_id of the virtual-switch we just created
            sw_val[8] = newnodeoutput.json()['node_id']
            # Request to delete the temporary template
            requests.delete(gns3_url_in + 'templates/' + looped_template_id, timeout=30)

            # Change the name of the GNS3 node that we just created
            requests.put(gns3_url_in + 'projects/' + prj_id_in + '/nodes/' + sw_val[8],
                         timeout=30, json={'name': sw_val[0]})
            # Capture the docker_id of the virtual-switch we just created  (container
            # re-spawned when we changed its name)
            sw_val[9] = requests.get(gns3_url_in + 'projects/' + prj_id_in + '/nodes/' +
                                     sw_val[8], timeout=30).json()['properties']['container_id']
            # Copy the modified sw_val objects contents back into sw_vals_in[sw_val_ctr]
            sw_vals_in[sw_val_ctr] = sw_val
            # Increment x/y coordinates for the *next* switch to be instantiated
            nodex += 150
            if nodex > 400:
                nodex = -800
                nodey = nodey + 200
            # Tell GNS3 to start the node that represents the current switch
            requests.post(gns3_url_in + 'projects/' + prj_id_in + '/nodes/' + sw_val[8]
                          + '/start', timeout=30)
            # Rebuild the switch-config from its current state as a list of individual
            # lines to a single string with newline characters.
            my_string_to_go = ''
            for i in allconf_in[sw_val_ctr]:
                my_string_to_go = my_string_to_go + i + "\n"
            # Apply ASCII encoding to the config string
            ascii_to_go = my_string_to_go.encode('ascii')
            # Turn the ASCII-encoded string into a bytes-like object
            bytes_to_go = BytesIO(ascii_to_go)
            # file_like_to_go = StringIO(my_string_to_go)
            # Turn the switch-config string into a tar archive for later
            fh = BytesIO()
            with tarfile.open(fileobj=fh, mode='w') as tarch:
                info = tarfile.TarInfo('startup-config')
                info.size = len(my_string_to_go)
                tarch.addfile(info, bytes_to_go)
            # Get a docker API connection for the current switch's container
            cont1 = d_clnt.containers.get(sw_val[9])
            # Retrieve our tar archive from the file-like object ('fh') that we stored it in
            uggo = fh.getbuffer()
            # Put the startup-config onto / on the virtual-switch
            cont1.put_archive('/', data=uggo)
            # Move the startup-config from / to /mnt/flash on the virtual switch
            cont1.exec_run('mv /startup-config /mnt/flash/')
            # Set URL for request to GNS3 to stop the node
            url = gns3_url_in+"projects/" + prj_id_in+"/nodes/" + sw_val[8] + "/stop"
            # Assign the HTTP post request for async execution
            tasks.append(asyncio.ensure_future(gns3_post(session, url, 'post')))
            await asyncio.sleep(0.2)
        await asyncio.gather(*tasks)
        switch_vals_out = sw_vals_in
        return switch_vals_out


async def gns3_connx_create_async(servername_in: str, gns3_url_in: str, sw_vals_new:
                                  list, connx_in: list, prj_id_in: str):
    """Add nodes to the new GNS3 project and push a copy of the configuration files
    to their substrate docker containers.

    Parameters
    ----------
    servername_in : str
        The name of the aiohttp.ClientSession object used for the connections
    gns3_url_in : str
        The URL to be posted to the GNS3 server
    sw_vals_new : list
        List of needed data about the switches to be emulated.  GNS3 node IDs were
        added by the gns3_nodes_create function.
    connx_in : list
        List of node-to-node connections that we want to create in the GNS3 project
    """

    async with aiohttp.ClientSession() as sesh2:
        # Loop through connections_to_make and make the connections between switches
        print("Instantiating the connections between switches in the GNS3 \
              project (might take a minute):")
        cnx_urls = []
        cnx_json = []
        for n, val in enumerate(connx_in):
            a_node_id = []
            b_node_id = []
            for m, val2 in enumerate(sw_vals_new):
                if val[0] == val2[5]:
                    a_node_id = val2[8]
                if val[2] == val2[5]:
                    b_node_id = val2[8]
            a_node_adapter_nbr = str(val[1].split('/')[0].split('ethernet')[1])
            b_node_adapter_nbr = str(val[3].split('/')[0].split('ethernet')[1])
            # Make a list of URLs for the requests to create all the links
            cnx_urls.append('')
            cnx_urls[n] = gns3_url_in + 'projects/' + prj_id_in + '/links'
            cnx_json.append({})
            cnx_json[n] = {'nodes': [{'adapter_number': int(a_node_adapter_nbr),
                                      'node_id': a_node_id, 'port_number': 0},
                                     {'adapter_number': int(b_node_adapter_nbr),
                                      'node_id': b_node_id, 'port_number': 0}]}
        # Assign the HTTP post request for async execution
        tasks_2 = []
        for n, url in enumerate(cnx_urls):
            tasks_2.append(asyncio.ensure_future(gns3_post(sesh2, str(url),
                           'post', jsondata=cnx_json[n])))
            await asyncio.sleep(0.2)
        responses = await asyncio.gather(*tasks_2)
        return responses


async def gns3_post(session_in: str, url_in: str, method: str, **kwargs) -> str:
    """Send an async POST request to GNS3 server.

    Parameters
    ----------
    session_in : str
        The name of the aiohttp.ClientSession object used for the connections
    url_in : str
        The URL to be posted to the GNS3 server (includes project ID and node ID)
    method : str
        The HTTP method (get, put, post) to be used in the request
    jsondata : str
        Optional.  Any JSON to be included with the HTTP request
    """

    if 'jsondata' in kwargs:
        jsondata = kwargs['jsondata']
    if method == 'post':
        if 'jsondata' in kwargs:
            async with session_in.post(url_in, json=jsondata) as response:
                await asyncio.sleep(.2)
        else:
            async with session_in.post(url_in) as response:
                await asyncio.sleep(.2)
    if method == 'get':
        if 'jsondata' in kwargs:
            async with session_in.get(url_in, json=jsondata) as response:
                await asyncio.sleep(.2)
        else:
            async with session_in.get(url_in) as response:
                await asyncio.sleep(.2)
    if method == 'put':
        if jsondata:
            async with session_in.put(url_in, json=kwargs['jsondata']) as response:
                await asyncio.sleep(.2)
        else:
            async with session_in.put(url_in) as response:
                await asyncio.sleep(.2)
    return response



"""eos_sanitizer.py

For modifying the configuration extracted from an Arista hardware
 appliance's EOS configuration to make it suitable for use on a cEOS
 container in a lab environment"""


def eos_to_ceos(switchConfigIn: list, sysMacIn: str) -> list:
    """Module entry-point.  Accepts an Arista EOS switch config
     and related paraand returns, returning a cEOS lab-ready
     version of the configuration

    Parameters
    ---------
    switchConfigIn : list
        List of lines of a switch's configuration
    sysMacIn : str
        The system MAC address of the original switch"""

    # List of global-config commands that we should comment out for \
    # cEOS-compatibility and lab environments in general
    badStarts = ['radius',
                 'username',
                 'aaa',
                 'ip radius',
                 'hardware speed',
                 'queue',
                 'server ',
                 'ip radius',
                 'ntp server',
                 'daemon TerminAttr',
                 '   exec /usr/bin/TerminAttr']

    # Get the number of Ethernet interfaces present in the \
    # original config
    etherCount = count_ether_interfaces(switchConfigIn)

    # Replace all references to 'Management1' in the config with \
    # 'Ethernet0'
    #mgt_port_int = int(etherCount) + 1
    mgt_port_str = 'Management0'

    # Loop through the lines in each switch's configuration
    for linect, line in enumerate(switchConfigIn):
        # Replace the Management1 interface name with an extra Ethernet interface
        switchConfigIn[linect] = line.replace('Management1', mgt_port_str)
        switchConfigIn[linect] = line.replace('Management0', mgt_port_str)
        # Eiminate config lines the begin with any of the "badStarts" strings
        for oopsie in badStarts:
            if switchConfigIn[linect].startswith(oopsie):
                # Can't just delete the un-wanted lines, that would screw up
                # the iteration through the list. Better to just prepend with a '!'
                switchConfigIn[linect] = "!removed_for_cEOS-lab| " + switchConfigIn[linect]
        # Get rid of '...netN/2|3|4' interface config sections altogether
        # (can't have them getting converted to ../netN and their vestigial config
        # overwriting the actual interface config
        spurious_interface = False
        # Check to see if the current config line is a 'spurious' interface
        spurious_interface = switchConfigIn[linect].startswith(
            'interface Ethernet') and ('/2' in switchConfigIn[linect] or '/3'
                                       in switchConfigIn[linect] or '/4' in
                                       switchConfigIn[linect])
        if spurious_interface:
            # Loop through the lines in the spurious interface's config section
            # and comment them out by prepending with '!'
            next_sec = False
            shortcount = linect
            # Stop commenting out lines when we get to the end of the config
            # section (marked by a line consisting of '!')
            while not next_sec:
                if switchConfigIn[shortcount] == '!':
                    next_sec = True
                switchConfigIn[shortcount] = '!' + switchConfigIn[shortcount]
                shortcount += 1
        # Convert interface names from  '...netn/m' to '...netn'
        if switchConfigIn[linect].startswith('interface Ethernet'):
            switchConfigIn[linect] = switchConfigIn[linect].split('/')[0]

    # Add a configuration section to apply the system-mac-address of the
    # original switch to the cEOS container's configuration
    switchConfigIn = applySysMac(switchConfigIn, sysMacIn)
    return switchConfigIn, etherCount


def count_ether_interfaces(switchConfigIn: list) -> int:
    """Accept a list of lines representing a switch config and return \
    the number of Ethernet interfaces the corresponding cEOS \
    container will need

    Parameters
    ---------
    switchConfigIn : list
        List of lines of a switch's configuration

    Returns
    -------
    my_ethercount : int
        The number of Ethernet interfaces the cEOS container version of the switch will\
         need
    """
    my_ethercount = 0
    for line in switchConfigIn:
        # We're only counting single interfaces (not the breakout interfaces)
        if (line.startswith('interface Ethernet') and (not (line.endswith('/2') or
                                                            line.endswith('/3') or
                                                            line.endswith('/4')))):
            my_ethercount += 1
    return my_ethercount


def applySysMac(switchConfigIn: list, sysMacIn: str) -> list:
    """Construct and append the event handler configuration to to \
    apply the system-mac-address of the modeled switch to the cEOS
    container's configuration

    Parameters
    ---------
    switchConfigIn : list
        List of lines of a switch's configuration
    sysMacIn : str
        The system MAC address of the original switch

    Returns
    -------
    switchConfigIn : list
        List of lines of a switch's configuration
    """
    
    # Create an event-handler section to append to the configuration to
    # use the original switch's MAC address
    sysMacSnip = ['', '', '', '', '', '', '']
    sysMacSnip[0] = 'event-handler onStartup'
    sysMacSnip[1] = ' trigger on-boot'
    sysMacSnip[2] = ' action bash'
    sysMacSnip[4] = '  echo $var_sysmac > /mnt/flash/system_mac_address'
    sysMacSnip[5] = '  truncate -s -1 /mnt/flash/system_mac_address'
    sysMacSnip[6] = '  EOF'
    # Remove the last line ('end') of the config  and append the system_mac_config
    # snippet(with the REAL switch's MAC address) before adding the final 'end' back
    # This will help our lab switches look more like the prod switches, but will
    # also work around the system-mac MLAG bug on cEOS
    poppedline = switchConfigIn.pop(-1)
    sysMacSnip[3] = '      var_sysmac=\'' + \
        sysMacIn + '\''
    for sysmacline in range(len(sysMacSnip)):
        switchConfigIn.append(sysMacSnip[sysmacline])
    switchConfigIn.append(poppedline)
    return switchConfigIn



"""ptovnetlab module/script

Entry-point module for package.  Gathers and parses run-state details
 from a list of Arista network switches.  Then creates a GNS3 virtual-
 lab project in which the interrogated devices are emulated."""


import sys
from getpass import getpass
import requests
from ptovnetlab import arista_poller, arista_sanitizer, gns3_worker


def read_file(file_to_read: str) -> list:
    """Open a file and return its contents as a list of strings

    Parameters
    ----------
    file_to_read : str
        The path of the file to be read

    Returns
    -------
    list of lines : list
        The contents of the file as a list of strings
    """
    # Open the file in read mode
    opened_file = open(file_to_read, "r")
    # Each line of the file into an entry in a list called list_of_lines
    list_of_lines = opened_file.read().splitlines()
    # Close the file that was being read
    opened_file.close()
    return list_of_lines


def list_search(list_to_search: list, item_to_find: str) -> bool:
    """Search a list for a specified string

    Parameters
    ----------
    list_to_search : list
        The list to be searched
    item_to_find : str
        The item to search for
    """
    for val in list_to_search:
        if val == item_to_find:
            return True
    return False


def predelimiter(string, delimiter):
  """Returns the section of a string before the first instance of a delimiter."""

  index = string.find(delimiter)
  if index == -1:
    return string  # Delimiter not found, return the whole string
  else:
    return string[:index]


def p_to_v(**kwargs):
    """Pull switch run-state data, massage it, and turn it into a GNS3 lab

    Parameters
    ---------
    username : str
        The username to be used in authenticating eapi calls to the switches
    passwd : str
        The password to be used in authenticating eapi calls to the switches
    filename : str
        Name of the file containing the list of switches to be interrogated
    switchlist : list
        List object of names of switches to be interrogated.
    servername : str
        The name of the remote GNS3/Docker server
    prjname : str
        The name of the project to create on the GNS3 server

    Returns
    -------
    result : str
    """
    # Guide for which values are held in which indices of switch_vals
    # switch_vals[n][0]= switch name received as input argument
    # switch_vals[n][1]= switch model
    # switch_vals[n][2]= EOS version
    # switch_vals[n][3]= system MAC
    # switch_vals[n][4]= serial number
    # switch_vals[n][5]= LLDP system-name
    # switch_vals[n][6]= Number of Ethernet interfaces
    # switch_vals[n][7]= GNS3 image-template ID
    # switch_vals[n][8]= GNS3 node-ID
    # switch_vals[n][9]= Docker container ID
    # switch_vals[n][10]= switch vendor/platform
    # switch_vals[n][11]= QEMU VM ID

    # Initializing list of LLDP neighbor details we'll use to build a list of connections for GNS3
    connections_to_make = []

    # Initializing nested list for holding the configs of ALL the switches
    allconfigs = []

    # Initializing a list for holding a list of gns3 template_id values and corresponding EOS \
    # version values
    image_map = []
    # Set default values for all anticipated arguments
    filename = ''
    username = ''
    passwd = ''
    switchlist = []
    servername = ''
    prj_name = ''
    run_type = 'module'

    # Pull any expected (and received) arguments from kwargs into their recipient
    # objects
    if 'filename' in kwargs:
        filename = kwargs['filename']
    if 'username' in kwargs:
        username = kwargs['username']
    if 'passwd' in kwargs:
        passwd = kwargs['passwd']
    if 'switchlist' in kwargs:
        switchlist = kwargs['switchlist']
    if 'servername' in kwargs:
        servername = kwargs['servername']
    if 'prjname' in kwargs:
        prj_name = kwargs['prjname']
    if 'runtype' in kwargs:
        run_type = kwargs['runtype']
    if (filename == '' and switchlist == []):
        print("Enter a switch name and press enter.")
        print("(Press Enter without entering a name when done.")
        while True:
            line = input()
            if not line:
                break
            switchlist.append(line)
    if (filename != '') and (switchlist != []):
        print("Don't pass both filename_arg AND switchlist_arg; choose one or the other.")
        exit(1)
    if prj_name == '':
        prj_name = input('Enter a value to use for the GNS project name when modeling the \
                         production switches: ')
    if servername == '':
        servername = input('Enter the name of the GNS3 server: ')
    # Read the input switch-list file, if a filename was provided
    if filename != '':
        switchlist = read_file(filename)

    # Remove any blank entries from switchlist
    switchlist = list(filter(None, switchlist))
    # Prompt for switch/EOS credentials if none were provided
    if username == '':
        username = input('Enter username for Arista EOS login: ')
    if passwd == '':
        passwd = getpass('Enter password for Arista EOS login: ')

    # Call eos_poller.invoker to get runstate from the Arista switches
    switch_vals, connections_to_make, allconfigs = arista_poller.invoker(switchlist,
                                                                      username, passwd,
                                                                      run_type)

    # Loop through all those configs and clean them up for life in a virtual lab,
    # while also grabbing the interface count for later
    for i, val in enumerate(allconfigs):
        allconfigs[i], switch_vals[i][6] = arista_sanitizer.eos_to_ceos(val, switch_vals[i][3])

    # Create a list of the LLDP local-IDs used by our switches
    our_lldp_ids = []
    for val in switch_vals:
        our_lldp_ids.append(val[5])

    # Sanitize connections_to_make list; removing any entries in which either end
    # is NOT one of our switches  (we can't tell GNS3 to create a connection to a
    # node that doesn't exist in the project.)
    connections_to_make[:] = [connx for connx in connections_to_make if
                              list_search(our_lldp_ids, connx[0]) and
                              list_search(our_lldp_ids, connx[2])]

    # Remove A|B-inverted entries in connections_to_make
    # (connections are directionless; so A<>B is the same as B<>A)
    for i in connections_to_make:
        for j in connections_to_make:
            if i[2]+i[3] == j[0]+j[1]:
                connections_to_make.remove(j)

    # Clean up "management1" in the connections_to_make list (using eth0
    #  instead.  Prod Switch management1 interface is presented in cEOS CLI
    #  as Management0, and Docker presents it to the container as eth0, which
    # is how it presents in GNS3

    for i, val in enumerate(connections_to_make):
        connections_to_make[i][1] = val[1].lower()
        connections_to_make[i][3] = val[3].lower()
        if val[1].startswith('management'):
            for j, val2 in enumerate(switch_vals):
                if val[0] == val2[5]:
                    connections_to_make[i][1] = 'ethernet0'
        if val[3].startswith('management'):
            for j, val2 in enumerate(switch_vals):
                if val[2] == val2[5]:
                    connections_to_make[i][3] = 'ethernet0'

    # Set GNS3 URL
    gns3_url = 'http://'+servername+':3080/v2/'
    gns3_url_noapi = 'http://'+servername+':3080/static/web-ui/server/1/project/'

    # Get all of the docker image templates from the GNS3 server so we can figure out
    # which template_id value maps to a specific EOS version when we start building
    # our instances
    r = requests.get(gns3_url + 'templates', timeout=20)
    for x in r.json():
        if x['template_type'] == 'docker':
            image_map.append([x['template_id'], x['image']])

    # Loop through image_map while looping through switch_vals, looking for imag_map
    # names that are good matches for the EOS version data captured in switch_vals,
    #  and appending each entry in switch_status with the GNS3 template_id from the
    #  matching entry in image_map
    for i in range(len(switch_vals)):
        for j in range(len(image_map)):
            # strip any trailing garbage from the EOS version reported by the switch API
            fudgedupeosverion = ('ceos:' + predelimiter(switch_vals[i][2].lower(), '-'))
            if fudgedupeosverion == image_map[j][1].lower():
                switch_vals[i][7] = image_map[j][0]
    # create a new project with provided name and grab the project_id
    gnsprj_id = requests.post(gns3_url + 'projects', json={'name': prj_name},
                              timeout=20).json()['project_id']
    # Grab the templates object from the GNS server so we can crawl through it
    # templ_qry_resp = requests.get(gns3_url + 'templates')

    # Invoke function that handles node creation/configuration in GNS3 project
    gns3_worker.invoker(servername, gns3_url, switch_vals,
                        allconfigs, gnsprj_id, connections_to_make)
    # Done!

    # Close the GNS3 project
    requests.post(gns3_url + 'projects' + 'project_id' + 'close')
    return gns3_url_noapi + gnsprj_id


if __name__ == '__main__':
    kwdict = {}
    for arg in sys.argv[1:]:
        splarg = arg.split('=')
        if splarg[0] == 'switchlist':
            splargalt = []
            for swname in splarg[1].split():
                splargalt.append(swname)
            kwdict[splarg[0]] = splargalt
        else:
            kwdict[splarg[0]] = splarg[1]
    kwdict['runtype'] = 'script'
    import ptovnetlab.gns3_worker
    import ptovnetlab.arista_poller
    p_to_v(**kwdict)
